
/************************************************************************/
/*									*/
/*									*/
/*    *****			  *****					*/
/*	*****			*****					*/
/*	  *****		      *****					*/
/*	    *****	    *****					*/
/*  ***************	  ***************				*/
/*  *****************	*****************				*/
/*  ***************	  ***************				*/
/*	    *****	    *****	    TheNetNode			*/
/*	  *****		      *****	    (PC)			*/
/*	*****			*****	    Public Domain		*/
/*    *****			  *****	    NORD><LINK			*/
/*									*/
/* This software ist public domain ONLY for non commercial use		*/
/*									*/
/* RS232_PC.C	-	Timer Routinen NUR fuer PC !			*/
/*			DIESES MODUL IST HARDWAREABHAENGIG !		*/
/*									*/
/************************************************************************/

#include "all.h"
#include <stdio.h>
#include <dos.h>

/*----------------------------------------------------------------------*/
/* Routinen zur Verwaltung der RS232 Schnittstelle			*/
/* 									*/
/* Die Schnittstelle wird im Interrupt betrieben, wovon das aufrufende	*/
/* Programm aber nichts merkt. Ausgaben erfolgen ueber einen Ringpuffer	*/
/* und Eingaben werden ueber einen separaten Buffer gepuffert. Status	*/
/* liefert daher immer den Buffer Status, nicht den aktuellen Status	*/
/* der Hardware. Bufferueberlauf wird gemeldet, sonst wird nichts	*/
/* unternommen.								*/
/*									*/
/* 13-02-90, DF2AU							*/
/*									*/
/* 07-08-90, DF2AU: 2 Stop Bits						*/
/* 29-09-90, DF2AU: nur RTS toggle					*/
/* 28-10-90, DF2AU: Fehler in reset_rs232() beseitigt			*/
/* 01-01-91, DF2AU: Adresse COM1 wird aus Bios geholt			*/
/*           DF2AU: ((inportb(uart +5) & 0x20) == 0x20)(tx interr.)     */
/* 04-11-91, ON5ZS: interrupt function for 8250/16450 AND 16550 +       */
/*           	    max_buf_rx and max_buf_tx			        */
/*                  modifications in rs232_in,rs232_out,                */
/*                     init_rs232,clear_rs232,reset_rs232               */
/* 08-01-92, ON5ZS: setbaud for pc Xtal  1.8 et 4.9 mhz                 */
/* 10-jan-92 ON5ZS: correction bug in TX interrupt for 16550            */
/* 15-jan-92 ON5ZS: replace sebaud_#### by "char setbaud(baud_rate);"   */
/*----------------------------------------------------------------------*/

#define BUFFERSIZE	0x3000		/* Groesse jedes Buffers	*/
#define INT_NUMMER	4		/* Interrupt Leitung		*/

extern int      is_16550;               /* flag ON if uart is 16550     */

static char 	in_buffer[BUFFERSIZE];  /* Zeichen von RS232 an Progr.*/
static char 	out_buffer[BUFFERSIZE]; /* Zeichen von Progr. an RS232*/
unsigned int   	in_buffer_in;		/* einschreiben in in_buffer	*/
unsigned int   	in_buffer_out;		/* auslesen aus in_buffer	*/
unsigned int   	out_buffer_in;		/* einschreiben in out_buffer	*/
unsigned int   	out_buffer_out;		/* auslesen aus out_buffer	*/
int             out_aktiv = 0;          /* = 0 after interrupt & 0 tx byte*/

					/* Basisadresse der Hardware	*/
int             trx_adr ;               /* RX & TX Registre          */
int             ier_adr ;               /* interrupt Enable registre */
int             iir_adr ;               /* Interrupt ident. registre */
int             fcr_adr ;               /* Fifo  Control registre */
int             lcr_adr ;               /* Line  Control registre */
int             mcr_adr ;               /* Modem Control registre */
int             lsr_adr ;               /* Line  Status registre  */
int             msr_adr ;               /* Modem Status registree */


void interrupt 	(*old_int)();		/* alter Interrupt Pointer	*/

/*----------------------------------------------------------------------*/
/* Interrupt Service by ON5ZS   					*/
/*----------------------------------------------------------------------*/
void interrupt rs232_int()
  {
  unsigned char  iir;       /* Interrupt Indentification Registre */
  int i=16;

  if (is_16550 == 0)

  /************************************************************************
  * <<<<<<<<<<<<<< interrupt function for 8250 and 16450 >>>>>>>>>>>>>>>> *
  ************************************************************************/

    {while ((iir = inportb(iir_adr)) != 1)        /* while interrup pending */
      {switch (iir)
	  {case 0x04:
	    /**************************************************************
	     * interrupt type is - Receiver data available                *
	     * interrupt reset with : read rx buffer register             *
	     **************************************************************/
	    if ((inportb(lsr_adr) & 0x01) == 0x01)       /* is data ready ? */
	       {                             /* read data + clear interrupt */
		 in_buffer[in_buffer_in++] = inportb(trx_adr);
		 if (in_buffer_in == BUFFERSIZE) in_buffer_in = 0;
	       }
	    break;

	   case 0x02:
	    /**************************************************************
	     * interrupt type is - Transmitter Holding Registre Empty     *
	     *   source int. is - Transmitter Holding Registre Empty      *
	     * interrupt reset with : reading iir if source               *
	     *                        or write in tx holding register     *
	     **************************************************************/
	    if (out_buffer_in != out_buffer_out)
	       {if ((inportb(lsr_adr) & 0x20) == 0x20)
		   {outportb(trx_adr, out_buffer[out_buffer_out++]);
		    if (out_buffer_out == BUFFERSIZE) out_buffer_out = 0;
		   }
		}
	     else
		out_aktiv = 0;

	    break;

	   case 0x06:
	    /************************************************************
	     * interrupt type is - receiver line status                 *
	     *   source int is  - Overun error   or  - Parity erorr     *
	     *                  - framming error or  - Break interrupt  *
	     * interrupt reset with : reading the Line Status Register  *
	     ************************************************************/
	    inportb(lsr_adr);                /* reset this interupt */

	    break;

	   case 0:
	    /**************************************************************
	     * interrupt type is - Modem stauts  (normal never activ)     *
	     *     source int is - CTS or DSR or Ring or DCD              *
	     * interrupt reset with : reading Modem Status Register       *
	     **************************************************************/
	    inportb(msr_adr);                          /* reset interrupt */
	    break;

	  }   /* end switch  */
       }      /* end while   */
    }         /*  end if     */

  else
  /*************************************************************************
   * <<<<<<<<<<<<<<<<<<<<< interrupt function for 16550 >>>>>>>>>>>>>>>>>> *
   ************************************************************************/

    {while ((iir = inportb(iir_adr)) != 0xC1)    /* while interrup pending */
       {switch (iir)
	  {case 0xC4:
	   case 0xCC:
	    /**************************************************************
	     * interrupt type is - Receiver data available (0xC4)  or     *
	     *                   - Caractere timeout   (0xCC)             *
	     * interrupt reset with : read rx buffer register  Drops      *
	     *                       below the trigger Level (0xC4)       *
	     *                     : read all the caracter in rx buffer   *
	     **************************************************************/
						    /* while data ready ? */
	   while((inportb(lsr_adr) & 0x01) == 0x01)
	     {in_buffer[in_buffer_in++] = inportb(trx_adr);
	      if (in_buffer_in == BUFFERSIZE) in_buffer_in = 0;
	     }

	   break;


	   case 0xC2:
	    /**************************************************************
	     * interrupt type is - Transmitter Holding Registre Empty     *
	     *    source int. is - Transmitter Holding Registre Empty     *
	     * interrupt reset with : reading iir if source               *
	     *                        or write in tx holding register     *
	     **************************************************************/
	    if (out_buffer_in == out_buffer_out )
		out_aktiv = 0;

	    else
		{if ((inportb(lsr_adr) & 0x20) == 0x20)
		   {do
		      {outportb(trx_adr, out_buffer[out_buffer_out++]);
		       if (out_buffer_out == BUFFERSIZE) out_buffer_out = 0;
		      }
		    while((out_buffer_in != out_buffer_out) && (--i > 0));
		   }
		 }

	    break;


	   case 0xC6:
	    /************************************************************
	     * interrupt type is - receiver line status                 *
	     *    source int is  - Overun error   or  - Parity erorr    *
	     *                   - framming error or  - Break interrupt *
	     * interrupt reset with : reading the Line Status Register  *
	     ************************************************************/
	    inportb(lsr_adr);                /* clear this interupt */

	    break;

	    case 0xC0:
	    /************************************************************
	     * interrupt type is - Modem stauts  (normal never activ)   *
	     *    source int is - CTS or DSR or Ring or DCD             *
	     * interrupt reset with : reading Modem Status Register     *
	     ************************************************************/
	    inportb(msr_adr);                         /* reset interrupt */
	    break;

	  }    /* end switch   */
       }       /* end while  int   */
    }          /* end else 8250 16450 / 16550    */

  outportb(0x20, 0x20);                 /* Int. Controller wieder frei */
  }            /* end function */

/*----------------------------------------------------------------------*/
/* Resetten der Schnittstelle						*/
/* OHNE Baudrate setzen!						*/
/* Buffer initialisieren                           			*/
/* Reset auf TNCs geben							*/
/*----------------------------------------------------------------------*/
void clear_RS232()
 {
  disable();                    	/* Int vorsichtshalber abschalt.*/
    if (is_16550 ==1)
      {outportb(fcr_adr,0xC7 );       /* FIFO ON , reset rx and tx FIFO */
       outportb(fcr_adr,0xC1);              /* trigger level = 14 bytes */
      }


    do
      {inportb(trx_adr);}                       /* Datenregister leeren */
    while((inportb(lsr_adr) & 0x01) == 0x01);

    in_buffer_in = in_buffer_out = out_buffer_in = out_buffer_out = 0;
    out_aktiv = 0;

  enable();			                       /* int wieder an */
 }

/*----------------------------------------------------------------------*/
/* Initialisieren der Schnittstelle					*/
/* OHNE Baudrate setzen!						*/
/* Buffer initialisieren, Interrupt Vektoren setzen			*/
/*----------------------------------------------------------------------*/
void init_rs232()
 {
  trx_adr = * ((int far *) MK_FP(0x40, 0x00));

  ier_adr = trx_adr + 1;               /* interrupt Enable registre */
  iir_adr = trx_adr + 2;               /* Interrupt ident. registre */
  fcr_adr = trx_adr + 2;               /* Fifo  Control registre */
  lcr_adr = trx_adr + 3;               /* Line  Control registre */
  mcr_adr = trx_adr + 4;               /* Modem Control registre */
  lsr_adr = trx_adr + 5;               /* Line  Status registre  */
  msr_adr = trx_adr + 6;               /* Modem Status registree */

  outportb(fcr_adr ,0x01);                 /* enable FIFO mode (16550) */

  if ((inportb(fcr_adr) & 0xC0) == 0xC0)         /* test if FIFO enabled */
    {is_16550 = 1;
     printf("UART is a 16550\n");
     outportb(fcr_adr,0xC7);           /* FIFO ON, clear rx and tx FIFO */
     outportb(fcr_adr,0xC1);           /* trigger level : 14 bytes   */
    }
  else
    {is_16550 = 0;
     printf("UART is a 8250/16450\n");
    }

#ifdef XTAL49152
    printf("Version for PC UART Xtal 4.9152 Mhz\n");
#else
    printf("Version for PC UART Xtal 1.8432 Mhz (standard value)\n");
#endif

  do
     {inportb(trx_adr);}                       /* Datenregister leeren */
  while((inportb(lsr_adr) & 0x01) == 0x01);

  outportb(lcr_adr, 0x07);  /* 8bit + 2 stop */         /* DLAB auf 0 */

  outportb(ier_adr, 0x07);                     /* Interrupt: Rx, TX, RLS */
  outportb(mcr_adr, 0x08);                       /* Interrupt frei (out2)*/

  in_buffer_in = in_buffer_out = out_buffer_in = out_buffer_out = 0;
  out_aktiv = 0;

  disable();
    old_int = getvect(8 + INT_NUMMER);
    setvect((8 + INT_NUMMER), rs232_int);              /* Vektor setzen */
    outportb(0x21, inportb(0x21) & (~(1 << INT_NUMMER)));
  enable();			                      /* Interrupt frei */


 }

/*----------------------------------------------------------------------*/
/* Schnittstelle sperren						*/
/*----------------------------------------------------------------------*/
void reset_rs232()
 {
  disable();
    outportb(0x21, inportb(0x21) | (1 << INT_NUMMER));
    outportb(ier_adr, 0x00);                /* keine Interrupt Freigabe */
    outportb(mcr_adr, 0x08);                   /* Out2 sperrt Interrupt */
    if (is_16550 == 1) outportb(fcr_adr,0x00);              /* FIFO OFF */
    setvect((8 + INT_NUMMER), old_int);
  enable();
 }

/*----------------------------------------------------------------------*/
/* RTS Leitung aktivieren						*/
/*----------------------------------------------------------------------*/
void rts_off()
 {
  outportb(mcr_adr, 0x0a);                /* Out2 an: Interrupt Freigabe */
 }

/*----------------------------------------------------------------------*/
/* RTS Leitung desaktivieren						*/
/*----------------------------------------------------------------------*/
void rts_on()
 {
  outportb(mcr_adr, 0x08);                /* Out2 an: Interrupt Freigabe */
 }

/*----------------------------------------------------------------------*/
/* set baud-rate by ON5ZS                                               */
/*               9600 19200 38400 57600 115200  Xtal pc uart 1.8 Mhz    */
/*               9600 19200 38400 51200 76800 102400         4.9 Mhz    */
/* return 0 if bad baud-rate                                            */
/*        1 if correct baud rate                                        */
/*        2 if correct baud-rate but too fast for a normal SIO          */
/*----------------------------------------------------------------------*/
int setbaud (long argu)
{
  char lsb = 0x06 ;                     /* default baud-rate 19200 baud */
  int  good_baud = 1;                   /* good baud rate flag          */


#ifdef XTAL49152
    /**********************************************************/
    /*              quartz uart PC = 4.9152 Mhz               */
    /**********************************************************/
    if (argu == 9600L)                   { lsb = 0x20; }
     else
       {if (argu == 19200L)              { lsb = 0x10; }
	else
	  {if (argu == 38400L)           { lsb = 0x08; }
	   else
	     {if (argu == 51200L)        { lsb = 0x06; }
	      else
		{if (argu == 76800L)     { lsb = 0x04; }
		 else
		   {if (argu == 102400L) { lsb = 0x03; }
		    else
		      {good_baud = 0;
       }  }  }  }  }  }

#else
    /**********************************************************/
    /*     quartz uart pc = 1.8432 Mhz (normal value)         */
    /**********************************************************/
    if (argu == 9600L)                   { lsb = 0x0C; }
     else
       {if (argu == 19200L)              { lsb = 0x06; }
	else
	  {if (argu == 38400L)           { lsb = 0x03; }
	   else
	     {if (argu == 57600L)        { lsb = 0x02; }
	      else
		{if (argu == 115200L)    { lsb = 0x01; }
		 else
		   {good_baud = 0;
       }  }  }  }  }
#endif

  if (argu > 38400L && is_16550==0 && good_baud ==1) good_baud = 2;

  if (good_baud != 2)
   {outportb(lcr_adr, 0x87);    /* auf Baudraten Register  2 stop bit */

    outportb(trx_adr, lsb );    /* LSB Teilerfaktor             */
    outportb(ier_adr, 0x00);    /* MSB                          */

    outportb(lcr_adr, 0x07);    /* 8 Bit, keine Paritaet, 2 Stop*/
   }

   return (good_baud);
}

/*----------------------------------------------------------------------*/
/* Zeichen aus Buffer an Programm liefern				*/
/*----------------------------------------------------------------------*/
char rs232_in()
 {
  char	c;
					   /* attend le 1 ier caractere */
  while (in_buffer_in == in_buffer_out);    /* ggfs. warten auf Zeichen	*/

  c = in_buffer[in_buffer_out++];
  if (in_buffer_out == BUFFERSIZE) in_buffer_out = 0;

  return(c);
 }

/*----------------------------------------------------------------------*/
/* Zeichen an Schnittstelle geben, ggfs im Buffer speichern		*/
/*----------------------------------------------------------------------*/
void rs232_out(char c)
 {
  int next_buffer_in ;

  next_buffer_in = out_buffer_in + 1 ;
  if (next_buffer_in == BUFFERSIZE) next_buffer_in = 0;

  disable();

   if (out_aktiv == 1 )
     {out_buffer[out_buffer_in] = c;
      out_buffer_in = next_buffer_in;
      enable();
     }

   else
     { enable();
       out_aktiv = 1;
       outportb(trx_adr,c);
     }
  }

/*----------------------------------------------------------------------*/
/* Status Eingabe							*/
/*----------------------------------------------------------------------*/
int rs232_in_status()
 { return(!(in_buffer_in == in_buffer_out));}

/*----------------------------------------------------------------------*/
/* Ende von RS232_PC.C */




